@@ -1,5 +1,4 @@
 from collections import OrderedDict
-import os
 import re
 import pprint
 import sys
@@ -54,7 +53,7 @@ PER_MODULE_OPTIONS = {
 }  # type: Final
 
 OPTIONS_AFFECTING_CACHE = ((PER_MODULE_OPTIONS |
-                            {"platform", "bazel", "plugins", "new_semantic_analyzer"})
+                            {"platform", "bazel", "plugins"})
                            - {"debug_cache"})  # type: Final
 
 
@@ -86,14 +85,6 @@ class Options:
         # PEP 420 namespace packages
         self.namespace_packages = False
 
-        # Use the new semantic analyzer
-        new_analyzer = os.getenv('NEWSEMANAL')
-        if new_analyzer:
-            # Use NEWSEMANAL=0 to change the default (for tests).
-            self.new_semantic_analyzer = bool(int(new_analyzer))
-        else:
-            self.new_semantic_analyzer = True
-
         # disallow_any options
         self.disallow_any_generics = False
         self.disallow_any_unimported = False
@@ -147,6 +138,10 @@ class Options:
         # Show "note: In function "foo":" messages.
         self.show_error_context = False
 
+        # Use nicer output (when possible).
+        self.color_output = True
+        self.error_summary = True
+
         # Files in which to allow strict-Optional related errors
         # TODO: Kill this in favor of show_none_errors
         self.strict_optional_whitelist = None   # type: Optional[List[str]]
@@ -216,6 +211,12 @@ class Options:
         # in modules being compiled. Not in the config file or command line.
         self.mypyc = False
 
+        # Disable the memory optimization of freeing ASTs when
+        # possible. This isn't exposed as a command line option
+        # because it is intended for software integrating with
+        # mypy. (Like mypyc.)
+        self.preserve_asts = False
+
         # Paths of user plugins
         self.plugins = []  # type: List[str]
 
@@ -242,6 +243,9 @@ class Options:
         # -- experimental options --
         self.shadow_file = None  # type: Optional[List[List[str]]]
         self.show_column_numbers = False  # type: bool
+        self.show_error_codes = False
+        # Use soft word wrap and show trimmed source snippets with error location markers.
+        self.pretty = False
         self.dump_graph = False
         self.dump_deps = False
         self.logical_deps = False
@@ -258,6 +262,11 @@ class Options:
         # Don't properly free objects on exit, just kill the current process.
         self.fast_exit = False
 
+    # To avoid breaking plugin compatability, keep providing new_semantic_analyzer
+    @property
+    def new_semantic_analyzer(self) -> bool:
+        return True
+
     def snapshot(self) -> object:
         """Produce a comparable snapshot of this Option"""
         # Under mypyc, we don't have a __dict__, so we need to do worse things.
